# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse import translate
from raytraverse.integrator.integrator import Integrator
import raytraverse.integrator._helpers as intg
from raytraverse.lightpoint import LightPointKD


def _prep_ds(lpts, skyvecs):
    try:
        snp = lpts[2]
    except IndexError:
        # this implies the sunpt is not needed (0 direct solar)
        return lpts[0:1], skyvecs[0:1], None
    skyvecs = [np.hstack(skyvecs[0:2]), skyvecs[2]]
    ski, skyvecs, refl = intg.apply_dsky_patch(lpts[0], lpts[1], skyvecs,
                                               snp.srcdir)
    lpts = [ski, snp]
    return lpts, skyvecs, refl


def evaluate_pt_ds(lpts, skyvecs, suns, **kwargs):
    lpts, skyvecs, refl = _prep_ds(lpts, skyvecs)
    return intg.evaluate_pt(lpts, skyvecs, suns, refl=refl, **kwargs)


def img_pt_ds(lpts, skyvecs, suns, **kwargs):
    lpts, skyvecs, refl = _prep_ds(lpts, skyvecs)
    return intg.img_pt(lpts, skyvecs, suns, refl=refl, **kwargs)


class IntegratorDS(Integrator):
    """specialized integrator for 2-phase DDS style calculation. assumes
    first lightplane is sky contrribution, second, direct sky contribution
    (with identical sampling to sky) and third direct sun contribution. Uses
    special point functions that combine two sky functions on a per patch basis.

    Parameters
    ----------
    skplane: raytraverse.lightfield.LightPlaneKD
    snplane: raytraverse.lightfiled.SunsPlaneKD
    dskplane: raytraverse.lightfield.LightPlaneKD
    """
    evaluate_pt = evaluate_pt_ds
    img_pt = img_pt_ds

    def __init__(self, skplane, dskplane, snplane, sunviewengine=None):
        super(IntegratorDS, self).__init__(skplane, dskplane, snplane,
                                           sunviewengine=sunviewengine)

    def _unroll_sky_grid(self, skydata, oshape):
        """

        Parameters
        ----------
        skydata: raytraverse.sky.SkyData
        oshape

        Returns
        -------

        """
        # broadcast skydata to match full indexing
        s = (*oshape[0:2], skydata.smtx.shape[1])
        smtx = np.broadcast_to(skydata.smtx[:, None, :], s).reshape(-1, s[-1])
        ds = (*oshape[0:2], skydata.sun.shape[1])
        dsns = np.broadcast_to(skydata.sun[:, None, :], ds).reshape(-1, ds[-1])
        skydatas = [smtx, dsns[:, 4:], dsns[:, 3:4]]
        return skydatas, dsns




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""parallelization functions for integration"""
import numpy as np
from scipy.spatial import Voronoi
from shapely.geometry import Polygon

from raytraverse import io, translate
from raytraverse.mapper import ViewMapper
from raytraverse.sampler import SunSamplerPtView
from raytraverse.lightpoint import LightPointKD


def evaluate_pt(lpts, skyvecs, suns, vm=None, vms=None,
                metricclass=None, metrics=None, srconly=False,
                sumsafe=False, suntol=1.0, svengine=None, blursun=False,
                refl=None, resamprad=0.0, **kwargs):
    """point by point evaluation suitable for submitting to ProcessPool"""
    if len(lpts) == 0:
        return np.zeros((len(suns), len(vms), len(metrics)))
    if srconly or sumsafe:
        sunskypt = lpts
        smtx = skyvecs
    else:
        lp0 = lpts[0]
        for lp in lpts[1:]:
            lp0 = lp0.add(lp)
        sunskypt = [lp0]
        smtx = [np.hstack(skyvecs)]

    if len(vms) == 1:
        args = (vms[0].dxyz, vms[0].viewangle*vms[0].aspect)
        didx = [lpt.query_ball(*args)[0] for lpt in sunskypt]
    else:
        didx = [None]*len(sunskypt)

    srcs = []
    resampi, resampvecs = prep_resamp(sunskypt, refl, resamprad)
    # loop over lightpoints to sum
    for ri, (lpt, di, sx) in enumerate(zip(sunskypt, didx, smtx)):
        pts = []
        t_srcview = (lpt.srcviews, lpt.srcviewidxs)
        # loop over skyvecs to evaluate
        for skyvec, sun in zip(sx, suns):
            if svengine is not None and ri == resampi:
                update_src_view(svengine, lpt, sun[0:3], vm, suntol, refl=refl,
                                resampvecs=resampvecs, resamprad=resamprad)
            vol = lpt.evaluate(skyvec, vm=vm, idx=di, blursun=blursun,
                               srconly=srconly)
            views = []
            for v in vms:
                views.append(metricclass(*vol, v, metricset=metrics,
                                         **kwargs)())
            views = np.stack(views)
            pts.append(views)
            lpt.set_srcviews(*t_srcview)
        srcs.append(np.stack(pts))
    return np.sum(srcs, axis=0)


def img_pt(lpts, skyvecs, suns, vms=None,  combos=None,
           qpts=None, skinfo=None, res=512, interp=False, prefix="img",
           suntol=1.0, svengine=None, refl=None, resamprad=0.0, **kwargs):
    """point by point evaluation suitable for submitting to ProcessPool"""
    outfs = []
    lpinfo = ["LPOINT{}= loc: ({:.3f}, {:.3f}, {:.3f}) src: ({:.3f}, {:.3f}, "
              "{:.3f}) {}".format(i, *lpt.pt, *lpt.srcdir[0], lpt.file)
              for i, lpt in enumerate(lpts)]
    if interp:
        lp0 = lpts[0]
        for lp in lpts[1:]:
            lp0 = lp0.add(lp)
        lpts = [lp0]
        skyvecs = [np.hstack(skyvecs)]

    vinfos = []
    for i, v in enumerate(vms):
        img, pdirs, mask, mask2, header = v.init_img(res)
        if interp == "high":
            lp_is, w = lpts[0].content_interp_wedge(svengine, pdirs[mask],
                                                    bandwidth=10, **kwargs)
            lp_is = (lp_is,)
        elif interp == "fast":
            lp_is, w = lpts[0].content_interp(svengine, pdirs[mask], **kwargs)
            lp_is = (lp_is,)
        elif interp:
            lp_is = [None] * len(lpts)
            w = None
        else:
            lp_is = [lpt.query_ray(pdirs[mask])[0] for lpt in lpts]
            w = None
        vinfos.append((i, v, pdirs, mask, lp_is, w))
    if interp in ['high', 'fast']:
        interp = 'precomp'
    resampi, resampvecs = prep_resamp(lpts, refl, resamprad)
    for j, (c, info, qpt, sun) in enumerate(zip(combos, skinfo, qpts, suns)):
        for i, v, pdirs, mask, lp_is, w in vinfos:
            header = [v.header(qpt), "SKYCOND= sunpos: ({:.3f}, {:.3f}, {:.3f})"
                      " dirnorm: {} diffhoriz: {}".format(*info)] + lpinfo
            for ri, (lp_i, lp, svec) in enumerate(zip(lp_is, lpts, skyvecs)):
                if svengine is not None and ri == resampi:
                    update_src_view(svengine, lp, sun[0:3], v, suntol,
                                    refl=refl, resampvecs=resampvecs, resamprad=resamprad)
                lp.add_to_img(img, pdirs[mask], mask, vm=v, interp=interp,
                              idx=lp_i, skyvec=svec[j], engine=svengine,
                              interpweights=w)
                if np.min(svec) < 0:
                    img = np.maximum(img, 0)
            outf = "{}_{}_{}_{}.hdr".format(prefix, *c[i])
            outfs.append(outf)
            io.array2hdr(img, outf, header)
            img[:] = 0
    return outfs


def prep_resamp(lpts, refl=None, resamprad=0.0):
    sunpt = [i for i, s in enumerate(lpts) if "_sun_" in s.src]
    sunpt2 = [i for i, s in enumerate(lpts) if "sun" in s.src]
    if len(sunpt) > 0:
        resampi = sunpt[0]
    elif len(sunpt2) > 0:
        resampi = sunpt2[0]
    else:
        resampi = -1
    srcdir = lpts[resampi].srcdir[-1:]
    if refl is not None:
        refl = translate.norm(refl.reshape(-1, 3))
        sunr = translate.reflect(srcdir, refl, False)
        srcdir = np.vstack((srcdir, sunr[0]))
    if resampi >= 0 and resamprad > 0.0:
        resampvecs = []
        for src in srcdir:
            i = lpts[resampi].query_ball(src, resamprad)
            resampvecs.append(i[0])
    else:
        resampvecs = None
    return resampi, resampvecs


def calc_omega(vecs, pm):
    """calculate area"""
    # border capture any infinite edges
    bordered = np.concatenate((vecs,
                               pm.bbox_vertices(pm.area**.5 * 10)))
    vor = Voronoi(bordered[:, 0:2])
    omega = []
    for i in range(len(vecs)):
        region = vor.regions[vor.point_region[i]]
        p = Polygon(vor.vertices[region])
        area = 0
        for bord in pm.borders():
            mask = Polygon(bord)
            area += p.intersection(mask).area
        omega.append(area)
    return np.asarray(omega)


def _in_view(vm, suna, sunb, tol=0.533):
    return (vm.degrees(suna)[0] <= (vm.viewangle + tol)/2 or
            vm.degrees(sunb)[0] <= (vm.viewangle + tol)/2)


def update_src_view(engine, lpt, sun, vm=None, tol=1.0, refl=None,
                    resampvecs=None, reflarea=None, resamprad=0.0):
    rtol = max(0.533, resamprad)
    if translate.degrees(sun, lpt.srcdir)[0] <= tol:
        return None
    if reflarea is None:
        reflarea = [0.533]
    if vm is None:
        vm = ViewMapper()
    inview = [_in_view(vm, sun, lpt.srcdir[-1], rtol)]
    svm = [ViewMapper(sun, reflarea[0], "sunview")]
    srcdir = [sun]
    if refl is not None and len(refl) > 0:
        if len(reflarea) < 1 + len(refl):
            reflarea += [0.533]*len(refl)
        refl = translate.norm(refl.reshape(-1, 3))
        sunr = translate.reflect(sun.reshape(1, 3), refl, False)
        for i, (sr, m, ar) in enumerate(zip(*sunr, reflarea[1:])):
            if m and ar > 0:
                rsrc = translate.reflect(lpt.srcdir[-1:], refl)[0][0]
                srcdir.append(rsrc)
                vm = ViewMapper(sr, ar, f"reflection_{i:03d}")
                svm.append(vm)
                inview.append(_in_view(vm, sr, rsrc, rtol))
    dosamp = np.any(inview)
    if len(svm) > 0 and dosamp:
        viewsampler = SunSamplerPtView(lpt.scene, engine, sun, 0)
        svpoints = viewsampler.run(lpt.pt, 0, vm=svm)
        lpt.set_srcviews(svpoints)
        if resampvecs is not None and dosamp:
            vecs = []
            ri = []
            for sd, rs, iv, sv in zip(srcdir, resampvecs, inview, svm):
                if inview and len(rs) > 0:
                    ymtx, pmtx = translate.rmtx_yp(np.stack((sd, sv.dxyz)))
                    # rotate to z-up in lp.srcdir space
                    vec = np.einsum("ij,kj,li->kl", ymtx[0], lpt.vec[rs], pmtx[0])
                    # rotate back in actual sun space
                    vec = np.einsum("ji,kj,il->kl", pmtx[1], vec, ymtx[1])
                    lucky_squirel = sv.degrees(vec) > 0.534/2
                    if np.sum(lucky_squirel) > 0:
                        vecs.append(vec[lucky_squirel])
                        ri.append(np.array(rs)[lucky_squirel])
            if len(ri) > 0:
                rvecs = np.vstack(vecs)
                ris = np.concatenate(ri)
                rvecs = np.concatenate(np.broadcast_arrays(lpt.pt[None, :],
                                                           rvecs), 1)
                lpt.lum[ris, -1] = engine(rvecs).ravel()


def apply_dsky_patch(skp, skd, skyvecs, skdir, dirlum=None):
    skpatch = translate.xyz2skybin(skdir, int((skp.srcn - 1)**.5))[0]
    skvec = skp.vec
    skydlum = np.copy(skd.lum[:, skpatch])
    sklum = np.maximum((skp.lum[:, skpatch] - skydlum), 0)[:, None]
    sklum = np.hstack((skp.lum, sklum))

    srcn = skp.srcn + 1
    if len(skyvecs[0]) < sklum.shape[1]:
        sklum = np.einsum('ij,kj->ki', skyvecs[0], sklum)
        srcn = len(skyvecs[0])
        skyvecs[0] = np.eye(len(skyvecs[0]))

    if dirlum is not None:
        srcn += 1
        sklum = np.hstack((sklum, dirlum))
        skyvecs = [np.hstack(skyvecs)]

    ski = LightPointKD(skp.scene, vec=skvec, lum=sklum, vm=skp.vm,
                       pt=skp.pt, posidx=skp.posidx,
                       src=f'sky_isun{skpatch:04d}', srcn=srcn,
                       srcdir=skdir,
                       write=False, omega=skp.omega, parent=skp.parent)
    try:
        refl = np.loadtxt(f"{skp.scene.outdir}/{skp.parent}/"
                          f"reflection_normals.txt")
    except OSError:
        refl = None
    else:
        refl = translate.norm(refl.reshape(-1, 3))
    return ski, skyvecs, refl



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================


"""factory functions for easy api access raytraverse."""
import os

import numpy as np

from raytraverse.integrator import ZonalIntegrator, ZonalIntegratorDS
from raytraverse.integrator import Integrator, IntegratorDS, IntegratorDV
from raytraverse.scene import Scene
from raytraverse.sky import SkyData
from raytraverse.mapper import PlanMapper
from raytraverse.lightfield import SunsPlaneKD, LightPlaneKD
from raytraverse.lightpoint import LightPointKD


def auto_reload(scndir, area, areaname="plan", skydata="skydata", ptres=1.0,
                rotation=0.0, zheight=None):
    """reload associated class instances from file paths

    Parameters
    ----------
    scndir: str
        matches outdir argument of Scene()
    area: str np.array
        radiance scene geometry defining a plane to sample, tsv file of
        points to generate bounding box, or np.array of points.
    areaname: str, optional
        matches name argument of PlanMapper()
    skydata: str, optional
        matches name argument of SkyData.write()
    ptres: float, optional
        resolution for considering points duplicates, border generation
        (1/2) and add_grid(). updateable
    rotation: float, optional
        positive Z rotation for point grid alignment
    zheight: float, optional
        override calculated zheight

    Returns
    -------
    Scene
    PlanMapper
    SkyData
    """
    if not os.path.exists(scndir):
        raise FileNotFoundError(f"auto_reload is only for reloading existing "
                                f"scenes. {scndir} does not exist")
    scn = Scene(scndir)
    if not os.path.isfile(f"{scndir}/{skydata}.npz"):
        raise FileNotFoundError(f"auto_reload is only for reloading existing "
                                f"skydata. {scndir}/{skydata}.npz does not "
                                f"exist")
    skd = SkyData(f"{scndir}/{skydata}.npz")
    pm = PlanMapper(area, name=areaname, ptres=ptres, rotation=rotation,
                    zheight=zheight)
    return scn, pm, skd


def load_lp(path, hasparent=True):
    if hasparent:
        ftree = path.rsplit("/", 3)
        scndir = ftree[-4]
        parent = ftree[-3]
    else:
        ftree = path.rsplit("/", 2)
        scndir = ftree[-3]
        parent = None
    scn = Scene(scndir)
    pidx = int(ftree[-1].split(".")[0])
    try:
        pts = np.loadtxt(path.replace(f"/{ftree[-1]}", "_points.tsv"))
    except FileNotFoundError:
        pt = (0, 0, 0)
    else:
        try:
            pt = pts[pidx, -3:]
        except IndexError:
            pt = pts[-3:]
    return LightPointKD(scn, parent=parent, src=ftree[-2], posidx=pidx, pt=pt)


stypes = ('1comp', '2comp', '3comp', '1compdv', 'directview', 'directpatch', 'sunonly',
          'sunpatch', 'skyonly')

stypedescriptions = {
    '1comp': "standard DC method, sky patch only, full contribution depending "
             "on skyengine settings",
    '2comp': "sky patch for sky contribution, sun run for sun contribution, "
             "depth of contributions depends on skyengine and sunengine "
             "settings, no approximation for sun from sky patch",
    '3comp': "2-phase DDS, sky handles sky+indirect sun, sun handles direct sun"
             " requires directskyrun -ab 1 and sunrun -ab 0",
    '1compdv': "standad DC method, but with direct view replacement of sun and"
               " specular reflections",
    'directview': "only evaluate srcviewpts (direct views to sun and specular "
                  "reflections",
    'directpatch': "only evaluate results from dskyrun",
    'sunonly': "only evaluate results from sunrun",
    'sunpatch': "use skyrun results to evaluate sun contribution",
    'skyonly': "use skyrun to evaluate sky contribution only"
    }

stypedocstring = "\n".join([f"    - {k}: {v}" for k,v in stypedescriptions.items()])


def get_integrator(scn, pm, srcname="suns", simtype="2comp", zonal=False,
                   sunviewengine=None):
    req_sun = ('2comp', '3comp', 'directview', 'sunonly')
    req_sky = ('1comp', '2comp', '3comp', 'sunpatch', 'skyonly', '1compdv')
    req_dsk = ('3comp', 'directpatch', '1compdv')
    sunfile = f"{scn.outdir}/{pm.name}/{srcname}_sunpositions.tsv"
    skpoints = f"{scn.outdir}/{pm.name}/sky_points.tsv"
    dskpoints = f"{scn.outdir}/{pm.name}/skydcomp_points.tsv"
    if zonal:
        itg = ZonalIntegrator
        itgds = ZonalIntegratorDS
        itgdv = IntegratorDV
    else:
        itg = Integrator
        itgds = IntegratorDS
        itgdv = IntegratorDV

    try:
        sunplane = SunsPlaneKD(scn, sunfile, pm, f"{srcname}_sun")
    except OSError:
        if simtype in req_sun:
            raise OSError(f"file: {sunfile} does not exist, make sure that a"
                          f" complete sun sampling exists")
        sunplane = None
    try:
        skyplane = LightPlaneKD(scn, skpoints, pm, "sky")
    except OSError:
        if simtype in req_sky:
            raise OSError(f"file: {skpoints} does not exist, make sure that a"
                          f" complete sky sampling exists")
        skyplane = None
    try:
        dskplane = LightPlaneKD(scn, dskpoints, pm, "skydcomp")
    except OSError:
        if simtype in req_dsk:
            raise OSError(f"file: {dskpoints} does not exist, make sure that a"
                          f" complete direct sky sampling exists")
        dskplane = None

    if simtype in ["1comp", "sunpatch", "skyonly"]:
        return itg(skyplane, includesky=simtype != "sunpatch",
                   includesun=simtype != "skyonly", sunviewengine=sunviewengine)
    if simtype == "2comp":
        return itg(skyplane, sunplane, sunviewengine=sunviewengine)
    if simtype == "3comp":
        return itgds(skyplane, dskplane, sunplane, sunviewengine=sunviewengine)
    if simtype == "1compdv":
        return itgdv(skyplane, dskplane, sunviewengine=sunviewengine)
    if simtype in ["directview", "sunonly"]:
        return itg(sunplane, includesky=False, sunviewengine=sunviewengine)
    if simtype == "directpatch":
        return itg(dskplane, includesky=False, sunviewengine=sunviewengine)
    raise ValueError(f"Error loading {simtype}")



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse.renderer.radiancerenderer import RadianceRenderer
from raytraverse.crenderer import cRcontrib
from raytraverse.formatter import RadianceFormatter as Fmt


class Rcontrib(RadianceRenderer):
    """singleton wrapper for c++ raytrraverse.crenderer.cRcontrib class

    this class sets default arguments, helps with initialization and setting
    cpu limits of the cRcontrib instance. see raytrraverse.crenderer.cRcontrib
    for more details.

    Parameters
    ----------
    rayargs: str, optional
        argument string (options and flags only) raises ValueError if arguments
        are not recognized by cRtrace.
    scene: str, optional
        path to octree
    nproc: int, optional
        if None, sets nproc to cpu count, or the RAYTRAVERSE_PROC_CAP
        environment variable
    skyres: float, optional
        approximate resolution for skypatch subdivision (in degrees). Patches
        will have (rounded) size skyres x skyres. So if skyres=10, each patch
        will be 100 sq. degrees (0.03046174197 steradians) and there will be
        18 * 18 = 324 sky patches.
    modname: str, optional
        passed the -m option of cRcontrib initialization
    ground: bool, optional
        if True include a ground source (included as a final bin)
    default_args: bool, optional
        if True, prepend default args to rayargs parameter

    Examples
    --------

    Basic Initialization and call::

        r = renderer.Rcontrib(args, scene)
        ans = r(vecs)
        # ans.shape -> (vecs.shape[0], 325)
    """
    name = 'rcontrib'
    #: raytraverse.crenderer.cRcontrib
    engine = cRcontrib
    ground = True
    side = 18
    srcn = 325
    modname = "skyglow"

    def __init__(self, rayargs=None, scene=None, nproc=None,
                 skyres=10.0, modname='skyglow', ground=True,
                 default_args=True):
        scene = self.setup(scene, ground, modname, skyres)
        super().__init__(rayargs, scene, nproc=nproc,
                         default_args=default_args)

    @classmethod
    def setup(cls, scene=None, ground=True, modname="skyglow", skyres=10.0):
        """set class attributes for proper argument initialization

        Parameters
        ----------
        scene: str, optional
            path to octree
        ground: bool, optional
            if True include a ground source (included as a final bin)
        modname: str, optional
            passed the -m option of cRcontrib initialization
        skyres: float, optional
            approximate resolution for skypatch subdivision (in degrees). Patches
            will have (rounded) size skyres x skyres. So if skyres=10, each patch
            will be 100 sq. degrees (0.03046174197 steradians) and there will be
            18 * 18 = 324 sky patches.

        Returns
        -------
        scene: str
            path to scene with added sky definition

        """
        cls.ground = ground
        if scene is not None:
            srcdef = Fmt.get_skydef((1, 1, 1), ground=ground, name=modname)
            scene = Fmt.add_source(scene, srcdef)
        cls.side = int(np.floor(90/skyres)*2)
        cls.srcn = cls.side**2 + ground
        cls.modname = modname
        return scene

    @classmethod
    def get_default_args(cls):
        """construct default arguments"""
        # return f"-ab 7 -ad 10 -as 0 -lw 1e-5 -st 0 -ss 16 -c {10*cls.srcn}"
        return (f"-u+ -ab 16 -av 0 0 0 -aa 0 -as 0 -dc 1 -dt 0 -lr -14 -ad "
                f"{50*cls.srcn} -lw {0.008/cls.srcn} -st 0 -ss 16 -c 1")

    @classmethod
    def set_args(cls, args, nproc=None):
        """prepare arguments to call engine instance initialization

        Parameters
        ----------
        args: str
            rendering options
        nproc: int, optional
            cpu limit

        """
        args = (f" -V+ {args} -w- -e 'side:{cls.side}' -f scbins.cal "
                f"-b bin -bn {cls.srcn} -m {cls.modname}")
        super().set_args(args, nproc)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

class _VirtEngine:

    @classmethod
    def __call__(cls, rays):
        return None


class Renderer(object):
    """virtual singleton renderer class.
    the Renderer is implemented as a singleton as specific subclasses (rtrace,
    rcontrib) have many global variables set at import time. This ensures the
    python object is connected to the current state of the engine c++-class.

    All renderer classes are callable with with a numpy array of shape (N,6)
    representing the origin and direction of ray samples to calculate.
    """

    args = None
    _pyinstance = None
    instance = _VirtEngine()
    scene = None

    def __new__(cls, rayargs=None, scene=None, nproc=None, **kwargs):
        if cls._pyinstance is None:
            cls._pyinstance = object.__new__(cls)
        return cls._pyinstance

    @classmethod
    def __call__(cls, rays):
        return cls.instance(rays)

    @classmethod
    def set_args(cls, args, nproc=None):
        pass

    def run(self, *args, **kwargs):
        """alias for call, for consistency with SamplerPt classes for nested
        dimensions of evaluation"""
        return self(args[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.craytraverse"""
import os
import shutil
import sys

from raytraverse import renderer
from raytraverse.sampler import draw
from raytraverse.formatter import RadianceFormatter
import numpy as np
from scipy import stats
import pytest


import clasp.script_tools as cst


@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    data = str(tmp_path_factory.mktemp("data"))
    shutil.copytree('tests/craytraverse/', data + '/test')
    cpath = os.getcwd()
    # use temp
    path = data + '/test'
    # uncomment to use actual (to debug results)
    # path = cpath + '/tests/craytraverse'
    os.chdir(path)
    yield path
    os.chdir(cpath)


def test_from_pdf():
    rv = stats.norm()
    nsamp = 500000
    t = .01
    x = np.linspace(0, 5, nsamp)
    pdf = rv.pdf(x)
    exp = np.sum(pdf > t)
    # c2 = draw.from_pdf2(pdf, t)
    c = draw.from_pdf(pdf, t)
    # print(np.sum(pdf > t), c.size, c2.size)
    # hist = np.histogram(x[c], 50)
    # mplt.quick_scatter([hist[1][1:], [0, 0]], [hist[0], [0, 7000]])
    # print(exp, c.size, c2.size)
    assert np.isclose(c.size, exp)


def test_empty_reset(tmpdir):
    rt = renderer.Rtrace("", "sky.oct")
    args = ('-I+ -ab 2 -ad 60000 -as 30000 -lw 1e-7')
    rc = renderer.Rcontrib(args, "scene.oct")
    rc.reset()
    rc.reset()
    rt.reset()
    rt.reset()
    assert True

# breaks test/example foor soome reason (must be in c-code!!!)
# def test_rcontrib_call(capfd, tmpdir):
#     args = ('-V+ -I+ -ab 2 -ad 60000 -as 30000 -lw 1e-7 -e side:6'
#             ' -f scbins.cal -b bin -bn 36 -m skyglow ')
#     cargs = f"rcontrib -n 5 -h- {args}  sky.oct"
#     check = cst.pipeline([cargs], inp='rays2.txt',
#                          forceinpfile=True)
#     check = np.fromstring(check, sep=' ').reshape(-1, 36, 3)
#     check = np.einsum('ikj,j->ik', check, [47.435/179, 119.93/179, 11.635/179])
#     r = renderer.Rcontrib(rayargs='-I+ -ab 2 -ad 600 -as 300 -c 100 -lw 1e-5',
#                           skyres=30, ground=False, scene="sky.oct")
#     # r.set_args('-I+ -ab 2 -ad 600 -as 300 -c 100 -lw 1e-5')
#     # r.load_scene("sky.oct")
#     vecs = np.loadtxt('rays2.txt')
#     # try:
#     #     with capfd.disabled():
#     #         ans, b = r.call(vecs, 'rays2.txt')
#     # except AttributeError:
#     test = r(vecs)
#     # test = np.fromstring(ans, sep=' ').reshape(-1, 36)
#     assert np.allclose(check, test, atol=.03)
#     r.reset()


def test_rtrace_call(tmpdir):
    args = "-ab 1 -ar 600 -ad 2000 -aa .2 -as 1500 -I+"
    cargs = f"rtrace -h {args} -n 4 sky.oct"
    check = cst.pipeline([cargs], inp='rays.txt',
                         forceinpfile=True)
    check = np.fromstring(check, sep=' ').reshape(-1, 3)
    check2 = np.einsum('ij,j->i', check, [47.435/179, 119.93/179, 11.635/179])
    # first load
    r = renderer.Rtrace(args, "sky.oct", default_args=False)
    vecs = np.loadtxt('rays.txt')
    ans = r(vecs)
    assert np.allclose(check, ans, atol=.03)

    # change output
    r.update_ospec('ZL')
    ans = r(vecs)
    assert np.allclose(check2, ans[:, 0], atol=.03)
    #
    # reload and change output to float
    r.reset()
    args2 = args + ' -oZ'
    r = renderer.Rtrace(args, "sky.oct", default_args=True)
    # test3 = r(vecs).ravel()
    # assert np.allclose(check2, test3, atol=.03)
    #
    # reload new scene
    r.load_scene("sky.oct")
    test3 = r(vecs).ravel()
    assert np.allclose(check2, test3, atol=.03)
    # #
    # # change args
    r.set_args("-lr 0 -oZ -I")
    test3 = r(vecs).ravel()
    assert np.allclose([0, 0, 0, 0, np.pi*2], test3, atol=.03)
    # #
    # # change back
    r.set_args(args2)
    test3 = r(vecs).ravel()
    assert np.allclose(check2, test3, atol=.03)
    # #
    # load sources
    r.set_args("-lr 0")
    r.update_ospec("Z")
    r.load_scene("scene.oct")
    r.load_source("sun.rad")
    test = r(vecs).ravel()
    r.load_source("sun2.rad")
    test2 = r(vecs).ravel()
    assert np.allclose(test * 2, test2, atol=.03)
    cargs = f"rtrace -h -ab 0 -n 4 sun.oct"
    check = cst.pipeline([cargs], inp='rays.txt',
                         forceinpfile=True)
    check = np.fromstring(check, sep=' ').reshape(-1, 3)
    check2 = np.einsum('ij,j->i', check, [47.435/179, 119.93/179, 11.635/179])
    assert np.allclose(test, check2, atol=.03)
    #
    #
    r.set_args(args)
    test2 = r(vecs).ravel()
    r.load_source("sky.rad")
    test3 = r(vecs).ravel()
    r.load_source("sun2.rad", 0)
    test4 = r(vecs).ravel()
    r.load_source("sun.rad")
    test5 = r(vecs).ravel()
    assert np.allclose(test2 + test3, test4, atol=.03)
    assert np.allclose(test2, test5, atol=.03)
    r.reset()


def test_ambient_reset(tmpdir):

    formatter = RadianceFormatter()
    args = "-u- -ab 1 -ar 1000 -ad 4000 -aa .05 -as 2000 -I+"
    r = renderer.Rtrace(args, "scene.oct", default_args=False)
    vecs = np.loadtxt('rays2.txt')

    def load_sun(sun, val, af=None):
        srcdef = f'tmp_sun.rad'
        f = open(srcdef, 'w')
        f.write(formatter.get_sundef(sun, (val, val, val)))
        f.close()
        r.load_source(srcdef, ambfile=af)
        os.remove(srcdef)

    load_sun((0, -.5, 1), 1000000, "temp.amb")
    r(vecs)
    a1 = r(vecs).ravel()
    load_sun((0, -.5, 1), 1000000, "temp2.amb")
    r(vecs)
    a2 = r(vecs).ravel()

    load_sun((-.5, -.5, 1), 2000000, "temp3.amb")
    r(vecs)
    a3 = r(vecs).ravel()

    load_sun((0, -.5, 1), 1000000, "temp.amb")
    a1a = r(vecs).ravel()
    load_sun((0, -.5, 1), 1000000, "temp2.amb")
    a2a = r(vecs).ravel()
    load_sun((0, -.5, 1), 1000000, "temp4.amb")
    r(vecs)
    a4 = r(vecs).ravel()
    r.set_args(r.defaultargs + " -I+ -ab 1 -c 30")
    aa0 = r(np.repeat(vecs[0:1], 1000, 0)).ravel()
    d = stats.norm(loc=np.mean(aa0), scale=np.std(aa0))
    ks = stats.kstest(aa0, d.cdf)
    # bm = np.random.default_rng().normal(loc=np.mean(aa0),
    #                                     scale=np.std(aa0),
    #                                     size=10000)
    # ks2 = stats.kstest(bm, d.cdf)
    assert ks[0] < .08
    r.set_args(args)
    load_sun((-.5, -.5, 1), 2000000, "temp3.amb")
    a3a = r(vecs).ravel()
    assert(np.allclose(a1, a1a))
    assert (np.allclose(a2, a2a))
    assert (np.allclose(a3, a3a))
    assert 1e-8 < np.sum(np.abs(a2 - a4)) < .5
    load_sun((-.5, -.5, 1), 2000000)
    noamb = r(vecs).ravel()
    assert 1e-8 < np.sum(np.abs(noamb - a3)) < 1
    r.reset()


def test_ambient_nostore(tmpdir):

    formatter = RadianceFormatter()
    args = "-u- -ab 1 -ar 1000 -ad 4000 -aa .05 -as 2000 -I+"
    r = renderer.Rtrace(args, "scene.oct", default_args=False, nproc=1)
    vecs = np.loadtxt('rays2.txt')

    def load_sun(sun, val):
        srcdef = f'tmp_sun.rad'
        f = open(srcdef, 'w')
        f.write(formatter.get_sundef(sun, (val, val, val)))
        f.close()
        r.load_source(srcdef)
        os.remove(srcdef)

    load_sun((0, -.5, 1), 1000000)

    # ambient values are not shared
    r(vecs)
    an2 = r(vecs).ravel()
    an3 = r(vecs).ravel()
    assert 1e-8 < np.sum(np.abs(an2 - an3)) < .5

    # ambient values are shared (vectors repeated, 1 process)
    an4 = r(np.repeat(vecs, 2, 0)).reshape(-1, 2).T
    assert (np.allclose(an4[0], an4[1]))
    r.reset()



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import sys

import numpy as np

from raytraverse import translate
from raytraverse.evaluate import MetricSet
from raytraverse.lightfield.lightresult import ResultAxis, LightResult
import raytraverse.integrator._helpers as intg
from raytraverse.mapper import ViewMapper
from raytraverse.utility import pool_call


class Integrator(object):
    """collection of lightplanes with KDtree structure for sun position query

    Parameters
    ----------
    lightplanes: Sequence[raytraverse.lightfield.LightPlaneKD]
    """
    evaluate_pt = intg.evaluate_pt
    img_pt = intg.img_pt

    def __init__(self, *lightplanes, includesky=True, includesun=True,
                 sunviewengine=None):
        self.scene = lightplanes[0].scene
        self.lightplanes = lightplanes
        self.includesky = includesky
        self.includesun = includesun
        self._sunviewengine = sunviewengine

    def make_images(self, skydata, points, vm, viewangle=180., res=512,
                    interp=False, prefix="img", namebyindex=False, suntol=10.0,
                    blursun=False, resamprad=0.0):
        """see namebyindex for file naming conventions

        Parameters
        ----------
        skydata: raytraverse.sky.Skydata
        points: np.array
            shape (N, 3)
        vm: Union[raytraverse.mapper.ViewMapper, np.array]
            either a predefined ViewMapper (used for all points) or an array
            of view directions (will use a 180 degree view angle when
            initializing ViewMapper)
        viewangle: float, optional
            view opening for sensor (0-180,360) when vm is given as an array
            of view directions.
        res: int, optional
            image resolution
        interp: bool, optional
            interpolate image
        prefix: str, optional
            prefix for output file naming
        namebyindex: bool, optional
            if False (default), names images by:
            <prefix>_sky-<row>_pt-<x>_<y>_<z>_vd-<dx>_<dy>_<dz>.hdr
            if True, names images by:
            <prefix>_sky-<row>_pt-<pidx>_vd-<vidx>.hdr, where pidx, vidx are
            refer to the order of points, and vm.

        Returns
        -------
        np.array of out_files shape (skies, points, views)

        """
        points = np.atleast_2d(points)
        if hasattr(vm, "dxyz"):
            vms = [vm]
        else:
            dxyz = np.asarray(vm).reshape(-1, 3)
            vms = [ViewMapper(d, viewangle) for d in dxyz]
        oshape = (len(skydata.maskindices), len(points), len(vms))
        idxs, skydatas, dsns, vecs = self._group_query(skydata, points)
        self.scene.log(self, f"Making Images for {oshape[2]}"
                             f" view directions at {oshape[1]} points under "
                             f"{oshape[0]} skies", True)
        skylabs = [f"sky-{si:04d}" for si in skydata.maskindices]
        if namebyindex:
            ptlabs = [f"pt-{i:03d}" for i in range(len(points))]
            viewlabs = [f"vd-{i:03d}" for i in range(len(vms))]
        else:
            ptlabs = ["pt-{:.1f}_{:.1f}_{:.1f}".format(*pt) for pt in points]
            viewlabs = ["vd-{:.1f}_{:.1f}_{:.1f}".format(*v.dxyz) for v in vms]
        a, b, c = np.meshgrid(skylabs, ptlabs,
                              viewlabs, indexing='ij')
        combos = np.stack((a.ravel().astype(object), b.ravel(), c.ravel()))
        combos = combos.T.reshape(-1, len(vms), 3)

        # use parallel processing
        skinfo = skydata.skydata[skydata.fullmask]
        s = (*oshape[0:2], skinfo.shape[1])
        skinfo = np.broadcast_to(skinfo[:, None, :], s).reshape(-1, s[-1])
        mask_kwargs = dict(combos=combos, qpts=vecs[:, 3:], skinfo=skinfo)
        outfs, _ = self._process_mgr(idxs, skydatas, dsns, type(self).img_pt,
                                     message="Making Images",
                                     mask_kwargs=mask_kwargs, vms=vms, res=res,
                                     interp=interp, prefix=prefix,
                                     suntol=suntol, blursun=blursun,
                                     resamprad=resamprad)
        return sorted([i for j in outfs for i in j])

    def evaluate(self, skydata, points, vm, viewangle=180.,
                 metricclass=MetricSet, metrics=None, datainfo=False,
                 srconly=False, suntol=10.0, blursun=False, coercesumsafe=False,
                 **kwargs):
        """apply sky data and view queries to daylightplane to return metrics
        parallelizes and optimizes run order.

        Parameters
        ----------
        skydata: raytraverse.sky.Skydata
        points: np.array
            shape (N, 3)
        vm: Union[raytraverse.mapper.ViewMapper, np.array]
            either a predefined ViewMapper (used for all points) or an array
            of view directions (will use 'viewangle' when
            initializing ViewMapper)
        viewangle: float, optional
            view opening for sensor (0-180,360) when vm is given as an array
            of view directions, note that for illuminance based
            metrics, a value of 360 may not make sense as values
            behind will be negative.
        metricclass: raytraverse.evaluate.BaseMetricSet, optional
        metrics: Sized, optional
        srconly: bool, optional
            sun only calculations
        suntol: float, optional
            if Integrator has an engine, resample sun views when actual sun
            position error is greater than this many degrees.
        blursun: bool, optional
            apply human PSF to small bright sources
        coercesumsafe: bool, optional
            attempt to calculate sumsafe metrics
        datainfo: Union[Sized[str], bool], optional
            include information about source data as additional metrics. Valid
            values include: ["pt_err", "pt_idx", "src_err", "src_idx"].
            If True, includes all.

        Returns
        -------
        raytraverse.lightfield.LightResult
        """
        points = np.atleast_2d(points)
        (vm, vms, cmetrics, ometrics,
         sumsafe, needs_post) = self._check_params(vm, viewangle, metrics,
                                                   metricclass, coercesumsafe)
        tidxs, skydatas, dsns, vecs = self._group_query(skydata, points)
        oshape = (len(skydata.maskindices), len(points), len(vms), len(cmetrics))
        self.scene.log(self, f"Evaluating {oshape[3]} metrics for {oshape[2]}"
                             f" view directions at {oshape[1]} points under "
                             f"{oshape[0]} skies", True)
        fields, isort = self._process_mgr(tidxs, skydatas, dsns,
                                          type(self).evaluate_pt,
                                          message="Evaluating Points",
                                          srconly=srconly, sumsafe=sumsafe,
                                          metricclass=metricclass,
                                          metrics=cmetrics, vm=vm, vms=vms,
                                          suntol=suntol, blursun=blursun,
                                          **kwargs)
        # sort back to original order and reshape
        fields = np.concatenate(fields, axis=0)[isort].reshape(oshape)
        if needs_post:
            ofields = np.zeros(fields.shape[:-1] + (len(ometrics),))
            for i, m in enumerate(ometrics):
                if m in cmetrics:
                    ofields[..., i] = fields[..., cmetrics.index(m)]
                elif m == "dgp":
                    illum = fields[..., cmetrics.index("illum")]
                    pwsl2 = fields[..., cmetrics.index("pwsl2")]
                    t1 = 5.87 * 10**-5 * illum
                    t2 = 9.18 * 10**-2 * np.log10(1 + pwsl2/np.power(illum, 1.87))
                    ll = 1
                    if "lowlight" in kwargs and kwargs["lowlight"]:
                        ll = np.where(illum < 500, np.exp(0.024*illum - 4) /
                                      (1 + np.exp(0.024*illum - 4)), 1)
                    ofields[..., i] = np.minimum(ll*(t1 + t2 + 0.16), 1.0)
                elif m == "ugp":
                    pwsl2 = fields[..., cmetrics.index("pwsl2")]
                    backlum = fields[..., cmetrics.index("backlum")]
                    with np.errstate(divide='ignore'):
                        ugr = np.maximum(0, 8*np.log10( 0.25*pwsl2/backlum))
                    ofields[..., i] = (1 + 2/7*10**(-(ugr + 5)/40))**-10
            fields = ofields
        sinfo, dinfo = self._sinfo(datainfo, vecs, tidxs, oshape[0:2])
        if sinfo is not None:
            nshape = list(sinfo.shape)
            nshape[2] = fields.shape[2]
            sinfo = np.broadcast_to(sinfo, nshape)
            fields = np.concatenate((fields, sinfo), axis=-1)
        # compose axes: (skyaxis, ptaxis, viewaxis, metricaxis)
        axes = (ResultAxis(skydata.rowlabel[skydata.fullmask], f"sky"),
                ResultAxis(points, "point"),
                ResultAxis([v.dxyz for v in vms], "view"),
                ResultAxis(list(ometrics) + dinfo, "metric"))
        lr = LightResult(fields, *axes)
        return lr

    @staticmethod
    def _check_params(vm, viewangle=180., metrics=None, metricclass=MetricSet,
                      coercesumsafe=False):
        if hasattr(vm, "dxyz"):
            vms = [vm]
        else:
            dxyz = np.asarray(vm).reshape(-1, 3)
            vm = ViewMapper()
            vms = [ViewMapper(d, viewangle) for d in dxyz]
        if metrics is None:
            metrics = metricclass.defaultmetrics
        needs_post = False
        if coercesumsafe:
            ometrics = []
            cmetrics = set()
            s2s = True
            for m in metrics:
                if m in metricclass.safe2sum:
                    cmetrics.add(m)
                    ometrics.append(m)
                elif m == "dgp":
                    cmetrics.update(["illum", "pwsl2"])
                    ometrics.append(m)
                    needs_post = True
                elif m == "ugp":
                    cmetrics.update(["backlum", "pwsl2"])
                    ometrics.append(m)
                else:
                    print(f"could not coerce metric {m} to sumsafe",
                          file=sys.stderr)
            cmetrics = list(cmetrics)
        else:
            ometrics = metrics
            cmetrics = metrics
            s2s = metricclass.check_safe2sum(metrics)
        return vm, vms, cmetrics, ometrics, s2s, needs_post

    def _group_query(self, skydata, points):
        # query and group sun positions
        gshape = (len(skydata.maskindices), len(points))
        suns = skydata.sun[:, None, 0:3]
        pts = np.atleast_2d(points)[None, :]
        vecs = np.concatenate(np.broadcast_arrays(suns, pts), -1).reshape(-1, 6)
        skydatas, dsns = self._unroll_sky_grid(skydata, gshape)
        idxs = []
        for lp in self.lightplanes:
            if lp.vecs.shape[1] == 6:
                idxs.append(lp.query(vecs)[0])
            else:
                idx, err = lp.query(points)
                idxs.append(np.broadcast_to(idx, gshape).ravel())
        return np.stack(idxs), skydatas, dsns, vecs

    def _sinfo(self, datainfo, vecs, idxs, oshape):
        """error and bin information for evaluate queries"""
        dinfo = ["src_err", "src_idx", "pt_err", "pt_idx"]
        if not datainfo:
            dinfo = []
        elif hasattr(datainfo, "__len__"):
            dinfo = [i for i in dinfo if i in datainfo]
        if len(dinfo) == 0:
            return None, dinfo
        dinfo2 = []
        sinfo = []
        for lp, idx in zip(self.lightplanes, idxs):
            for di in dinfo:
                if di == "src_err" and lp.vecs.shape[1] == 6:
                    snerr = np.linalg.norm(vecs[:, :3] - lp.vecs[idx, :3],
                                           axis=-1)
                    sinfo.append(translate.chord2theta(snerr)*(180/np.pi))
                    dinfo2.append(f"{lp.src}_{di}")
                elif di == "src_idx" and lp.vecs.shape[1] == 6:
                    srcidx = lp.data.idx[idx, 0]
                    sinfo.append(srcidx)
                    dinfo2.append(f"{lp.src}_{di}")
                elif di == "pt_err":
                    pterr = np.linalg.norm(vecs[:, 3:] - lp.vecs[idx, -3:],
                                           axis=-1)
                    sinfo.append(pterr)
                    dinfo2.append(f"{lp.src}_{di}")
                elif di == "pt_idx":
                    try:
                        ptidx = lp.data.idx[idx, 1]
                    except IndexError:
                        ptidx = idx
                    sinfo.append(ptidx)
                    dinfo2.append(f"{lp.src}_{di}")
        return np.array(sinfo).T.reshape(*oshape, 1, -1), dinfo2

    def _unroll_sky_grid(self, skydata, oshape):
        # broadcast skydata to match full indexing
        s = (*oshape, skydata.smtx.shape[1])
        smtx = np.broadcast_to(skydata.smtx[:, None, :], s).reshape(-1, s[-1])
        ds = (*oshape, skydata.sun.shape[1])
        dsns = np.broadcast_to(skydata.sun[:, None, :], ds).reshape(-1, ds[-1])

        skydatas = []
        # generic, assumes sun and/or sky, uses patch sun when no sun present
        skyi = None
        suni = None
        for i, lp in enumerate(self.lightplanes):
            if lp.vecs.shape[1] == 3:
                skyi = i
                skydatas.append(smtx)
            elif self.includesun:
                suni = i
                skydatas.append(dsns[:, 3:4])
        if self.includesun and suni is None and skyi is not None:
            dprx = skydata.smtx_patch_sun(includesky=self.includesky)
            dprx = np.broadcast_to(dprx[:, None, :], s).reshape(-1, s[-1])
            skydatas[skyi] = dprx
        return skydatas, dsns

    @staticmethod
    def _sort_run_data(tidxs):
        # unique returns sorted values (evaluation order)
        qtup, qidx = np.unique(tidxs.T, axis=0, return_index=True)
        tup_sort = np.lexsort(tidxs[::-1])
        # make an inverse sort to undo evaluation order
        tup_isort = np.argsort(tup_sort, kind='stable')
        return qtup, qidx, tup_isort

    def _process_mgr(self, tidxs, skydatas, dsns, eval_fn,
                     message="Evaluating Points", mask_kwargs=None,
                     **eval_kwargs):

        qtup, qidx, tup_isort = self._sort_run_data(tidxs)

        d = np.linspace(0, len(qtup), max(2, int(len(qtup)/250))).astype(int)
        slices = [slice(d[i], d[i + 1], 1) for i in range(len(d) - 1)]
        self.scene.log(self, f"Calculating {len(qidx)} sun/sky/pt combinations",
                       True)
        if self._sunviewengine is not None:
            eval_kwargs.update(svengine=self._sunviewengine)
        fields = []
        for i, slc in enumerate(slices):
            lptis = []
            for qi, qt in zip(qidx[slc], qtup[slc]):
                lptis.append(([(lp.data, tidx[qi]) for lp, tidx
                               in zip(self.lightplanes, tidxs)], qt))
            desc = f"{message} ({i+1:02d} of {len(slices):02d})"
            fields += pool_call(_load_pts, lptis, eval_fn, tidxs.T, mask_kwargs,
                                skydatas, dsns, desc=desc, pbar=self.scene.dolog, **eval_kwargs)
        return fields, tup_isort


def _load_pts(lpti, qt, eval_fn, idx_tup, mask_kwargs, skydatas, dsns,
              **kwargs):
    lpts = []
    sx = []
    mask = np.all(idx_tup == qt, -1)
    for (lp, tidx), smx in zip(lpti, skydatas):
        if tidx >= 0:
            lpts.append(lp[tidx])
            sx.append(smx[mask])
    if mask_kwargs is not None:
        for k, v in mask_kwargs.items():
            kwargs.update([(k, v[mask])])
    return eval_fn(lpts, sx, dsns[mask], **kwargs)



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""functions for translating from mappers to hdr"""
import numpy as np
import clasp.script_tools as cst


from raytraverse import translate, io
from raytraverse.evaluate import MetricSet, retina
from raytraverse.lightpoint import LightPointKD
from raytraverse.mapper.viewmapper import ViewMapper
from raytraverse.sampler import draw
from raytraverse.sampler.basesampler import filterdict


def uvarray2hdr(uvarray, imgf, header=None):
    res = uvarray.shape[0]
    vm = ViewMapper(viewangle=180)
    pixelxyz = vm.pixelrays(res)
    uv = vm.xyz2uv(pixelxyz.reshape(-1, 3))
    mask = vm.in_view(pixelxyz, indices=False)
    ij = translate.uv2ij(uv[mask], res)
    img = np.zeros(res*res)
    img[mask] = uvarray[ij[:, 0], ij[-1:None:-1, 1]]
    io.array2hdr(img.reshape(res, res), imgf, header)


def hdr2uvarray(imgf, vm=None, res=None):
    if vm is None:
        vm = hdr2vm(imgf)
    imarray = io.hdr2array(imgf)
    if res is None:
        res = imarray.shape[0]
    uv = translate.bin2uv(np.arange(res*res), res)
    xyz = vm.uv2xyz(uv)
    pxy = vm.ray2pixel(xyz, imarray.shape[0])
    return imarray[pxy[:, 1], pxy[:, 0]].reshape(res, res)


def hdr2vol(imgf, vm=None):
    ar = io.hdr2array(imgf).T
    if vm is None:
        vm = hdr2vm(imgf)
    vecs = vm.pixelrays(ar.shape[-1]).reshape(-1, 3)
    oga = vm.pixel2omega(vm.pixels(ar.shape[-1]), ar.shape[-1]).ravel()
    return vecs, oga, ar.ravel()


def hdr2vm(imgf, vpt=False):
    header = cst.pipeline([f"getinfo {imgf}"])
    if "VIEW= -vta" in header:
        vp = header.rsplit("VIEW= -vta", 1)[-1].splitlines()[0].split()
        view_angle = float(vp[vp.index("-vh") + 1])
        vd = vp.index("-vd")
        view_dir = [float(vp[i]) for i in range(vd + 1, vd + 4)]
        vpi = vp.index("-vp")
        view_pt = [float(vp[i]) for i in range(vpi + 1, vpi + 4)]
        hd = cst.pipeline([f"getinfo -d {imgf}"]).strip().split()
        x = 1
        y = 1
        for i in range(2, len(hd)):
            if 'X' in hd[i - 1]:
                x = float(hd[i])
            elif 'Y' in hd[i - 1]:
                y = float(hd[i])
        vm = ViewMapper(view_dir, view_angle * x / y)
    else:
        view_pt = None
        vm = None
    if vpt:
        return vm, view_pt
    else:
        return vm


def normalize_peak(v, o, l, scale=179, peaka=6.7967e-05, peakt=1e5, peakr=4,
                   blursun=False):
    pc = np.nonzero(l > peakt / scale)[0]
    if pc.size > 0:
        # first sort descending by luminance
        pc = pc[np.argsort(-l[pc])]
        pvol = np.hstack((v[pc], o[pc, None], l[pc, None]))
        # establish maximum radius for grouping
        cosrad = np.cos((peaka/np.pi)**.5*4)
        # calculate angular distance from peak ray and filter strays
        pd = np.einsum("i,ji->j", pvol[0, 0:3], pvol[:, 0:3])
        dm = pd > cosrad
        pc = pc[dm]
        pvol = pvol[dm]
        # calculate expected energy assuming full visibility:
        esun = pvol[0, 4]*peaka
        # sum up to peak energy
        cume = np.cumsum(pvol[:, 3]*pvol[:, 4])
        # treat as full sun
        if cume[-1] > esun:
            stop = np.argmax(cume > esun)
            if stop == 0:
                stop = len(cume)
            peakl = cume[stop - 1]/peaka
        # treat as partial sun (needs to use peak ratio)
        else:
            stop = np.argmax(pvol[:, 4] < pvol[0, 4]/peakr)
            if stop == 0:
                stop = len(cume)
            peakl = pvol[0, 4]
            peaka = cume[stop - 1]/peakl
        pc = pc[:stop]
        pvol = pvol[:stop]
        # new source vector weight by L*omega of source rarys
        pv = translate.norm(np.average(pvol[:, 0:3], axis=0,
                                       weights=pvol[:, 3]*pvol[:, 4]))
        # filter out source rays
        vol = np.delete(np.hstack((v, o[:, None], l[:, None])), pc, axis=0)
        v = np.vstack((vol[:, 0:3], pv))
        if blursun:
            cf = np.atleast_1d(retina.blur_sun(peaka, peakl))[0]
        else:
            cf = 1
        o = np.concatenate((vol[:, 3], [peaka*cf]))
        l = np.concatenate((vol[:, 4], [peakl/cf]))
    return v, o, l


def imgmetric(imgf, metrics, peakn=False, scale=179, threshold=2000., lowlight=False,
              **peakwargs):
    vm = hdr2vm(imgf)
    if vm is None:
        vm = ViewMapper(viewangle=180)
    v, o, l = hdr2vol(imgf, vm)
    if peakn:
        v, o, l = normalize_peak(v, o, l, scale, **peakwargs)
    return MetricSet(v, o, l, vm, metrics, scale=scale, threshold=threshold, lowlight=lowlight)()


def img2lf(imga, imgb, src, scn):

    accuracy = 0.5
    t0 = 0
    t1 = .6667
    levels = 6

    def _threshold(idx, acc):
        """threshold for determining sample count"""
        return acc * _linear(idx, t0, t1)

    def _linear(x, x1, x2):
        if levels <= 2:
            return (x1, x2)[x]
        else:
            return (x2 - x1)/(levels - 1) * x + x1

    vm, vp = hdr2vm(imga, vpt=True)
    uva = hdr2uvarray(imga, vm, 1024)
    if imgb is not None:
        vmb = ViewMapper(vm.dxyz*np.array((-1, -1, -1)), vm.viewangle)
        uvb = hdr2uvarray(imgb, vmb, 1024)
        uva = np.stack((uvb, uva), 0).reshape(-1, uvb.shape[1])
        vm = ViewMapper(vm.dxyz)
    accuracy *= np.average(uva)
    uvt = translate.resample(uva, uva.shape, radius=2)
    ar = int(uva.shape[0]/uva.shape[1])
    available = np.full(uvt.shape, True)
    rays = None
    vals = None
    for i in range(1, levels+1):
        res = 2**(levels-i)*1024/2**levels
        uvt = translate.resample(uvt, (res*ar, res))
        available = translate.resample(available.astype(float), uvt.shape) > 0.0
        p = draw.get_detail(uvt, *filterdict['wav']).reshape(uvt.shape)
        t = _threshold(levels-i, accuracy)
        p[np.logical_not(available)] = 0
        mi = p > t
        available[mi] = False
        miu = translate.resample(mi, (res*2*ar, res*2), False)
        uv = vm.idx2uv(np.arange(miu.size)[miu.ravel()], miu.shape, False)
        uv[:, 0] *= ar
        ray = vm.uv2xyz(uv)
        if rays is None:
            rays = ray
            vals = uva[miu]
        else:
            rays = np.concatenate((rays, ray))
            vals = np.concatenate((vals, uva[miu]))
        uva = translate.resample(uva, (res*ar, res))
    lp = LightPointKD(scn, rays, vals, vm, vp, src=src)
    lp.direct_view(512)






# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse import translate
from raytraverse.integrator.integrator import Integrator
import raytraverse.integrator._helpers as intg


def _prep_dv(lpts, skyvecs, skdir):
    dirlum = np.zeros((len(lpts[0].lum), 1))
    ski, skyvecs, refl = intg.apply_dsky_patch(lpts[0], lpts[1], skyvecs,
                                               skdir, dirlum)
    lpts = [ski]
    return lpts, skyvecs, refl


def evaluate_pt_dv(lpts, skyvecs, suns, **kwargs):
    side = int((skyvecs[0].shape[1] - 1)**.5)
    skpatch = translate.xyz2skybin(suns[0], side)[0]
    skdir = translate.skybin2xyz(skpatch, side)[0]
    lpts, skyvecs, refl = _prep_dv(lpts, skyvecs, skdir)
    return intg.evaluate_pt(lpts, skyvecs, suns, refl=refl, **kwargs)


def img_pt_dv(lpts, skyvecs, suns, **kwargs):
    side = int((skyvecs[0].shape[1] - 1)**.5)
    skpatch = translate.xyz2skybin(suns[0], side)[0]
    skdir = translate.skybin2xyz(skpatch, side)[0]
    lpts, skyvecs, refl = _prep_dv(lpts, skyvecs, skdir)
    return intg.img_pt(lpts, skyvecs, suns, refl=refl, **kwargs)


class IntegratorDV(Integrator):
    """specialized integrator for 2-phase Direct Views style calculation. assumes
    first lightplane is sky contrribution, second, direct sky contribution. Uses
    special point functions that combine two sky functions on a per patch basis.

    Parameters
    ----------
    skplane: raytraverse.lightfield.LightPlaneKD
    dskplane: raytraverse.lightfield.LightPlaneKD
    """
    evaluate_pt = evaluate_pt_dv
    img_pt = img_pt_dv

    def __init__(self, skplane, dskplane, sunviewengine):
        super(IntegratorDV, self).__init__(skplane, dskplane,
                                           sunviewengine=sunviewengine)

    def _group_query(self, skydata, points):
        # query and group sun positions
        gshape = (len(skydata.maskindices), len(points))
        suns = skydata.sun[:, None, 0:3]
        pts = np.atleast_2d(points)[None, :]
        vecs = np.concatenate(np.broadcast_arrays(suns, pts), -1).reshape(-1, 6)
        skydatas, dsns = self._unroll_sky_grid(skydata, gshape)
        idxs = []
        for lp in self.lightplanes:
            if lp.vecs.shape[1] == 6:
                idxs.append(lp.query(vecs)[0])
            else:
                idx, err = lp.query(points)
                idxs.append(np.broadcast_to(idx, gshape).ravel())
        idxs.append(np.broadcast_to(skydata.sunproxy[:, None], gshape).ravel())
        return np.stack(idxs), skydatas, dsns, vecs

    def _unroll_sky_grid(self, skydata, oshape):
        """

        Parameters
        ----------
        skydata: raytraverse.sky.SkyData
        oshape

        Returns
        -------

        """
        # broadcast skydata to match full indexing
        s = (*oshape[0:2], skydata.smtx.shape[1])
        smtx = np.broadcast_to(skydata.smtx[:, None, :], s).reshape(-1, s[-1])
        ds = (*oshape[0:2], skydata.sun.shape[1])
        dsns = np.broadcast_to(skydata.sun[:, None, :], ds).reshape(-1, ds[-1])
        # smtx (with sun), (patch sun, sun)
        skydatas = [np.hstack((smtx, dsns[:, 4:])), dsns[:, 3:4]]
        return skydatas, dsns



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski, HSLU and EPFL
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import pickle

import numpy as np
from scipy.interpolate import LinearNDInterpolator

from scipy.spatial import cKDTree, SphericalVoronoi, Voronoi
from shapely.geometry import Polygon
from clasp.script_tools import try_mkdir

from raytraverse import io, translate
from raytraverse.mapper import ViewMapper
from raytraverse.lightpoint.srcviewpoint import SrcViewPoint


class LightPointKD(object):
    """light distribution from a point with KDtree structure for directional
    query

    Parameters
    ----------
    scene: raytraverse.scene.BaseScene
    vec: np.array, optional
        shape (N, >=3) where last three columns are normalized direction vectors
        of samples. If not given, tries to load from scene.outdir
    lum: np.array, optional
        reshapeable to (N, srcn). sample values for each source corresponding
        to vec. If not given, tries to load from scene.outdir
    vm: raytraverse.mapper.ViewMapper, optional
        a default viewmapper for image and metric calculations, should match
        viewmapper of sampler.run() if possible.
    pt: tuple list np.array
        3 item point location of light distribution
    posidx: int, optional
        index position of point, will govern file naming so must be set to
        avoid clobbering writes. also used by spacemapper for planar sampling
    src: str, optional
        name of source group. will govern file naming so must be set to
        avoid clobbering writes.
    srcn: int, optional
        must match lum, does not need to be set if reloading from scene.outdir
    calcomega: bool, optional
        if True (default) calculate solid angle of rays. This  is
        unnecessary if point will be combined before calculating any metrics.
        setting to False will save some computation time.
    write: bool, optional
        whether to save ray data to disk.
    omega: np.array, optional
        provide precomputed omega values, if given, overrides calcomega
    """

    def __init__(self, scene, vec=None, lum=None, vm=None, pt=(0, 0, 0),
                 posidx=0, src='sky', srcn=1, srcdir=(0, 0, 1), calcomega=True,
                 write=True, omega=None, filterviews=True, srcviews=None,
                 parent=None, srcviewidxs=None):
        self.srcviews = []
        self.srcviewidxs = []
        self.set_srcviews(srcviews, srcviewidxs)
        if vm is None:
            vm = ViewMapper()
        #: raytraverse.mapper.ViewMapper
        self.vm = vm
        #: raytraverse.scene.Scene
        self.scene = scene
        #: int: index for point
        self.posidx = posidx
        #: np.array: point location
        self.pt = np.asarray(pt).flatten()[0:3]
        #: str: source key
        self.src = src
        #: direction to source(s)
        self.srcdir = translate.norm(np.asarray(srcdir).reshape(-1, 3))
        if parent is not None:
            outdir = f"{self.scene.outdir}/{parent}"
        else:
            outdir = self.scene.outdir
        self.parent = parent
        #: str: relative path to disk storage
        self.file = f"{outdir}/{self.src}/{self.posidx:06d}.rytpt"
        self._vec = np.empty((0, 3))
        self._lum = np.empty((0, srcn))
        self._d_kd = None
        self._omega = None
        if vec is not None and lum is not None:
            self.srcn = srcn
            self.update(vec, lum, omega=omega, calcomega=calcomega,
                        write=write, filterviews=filterviews)
        elif os.path.isfile(self.file):
            self.load()
        else:
            raise ValueError(f"Cannot initialize {type(self).__name__} without"
                             f" file: {self.file} or parameters 'vec' "
                             f"and 'lum'")
        self.srcdir = np.broadcast_to(self.srcdir, (self.srcn, 3))

    def load(self):
        f = open(self.file, 'rb')
        loads = pickle.load(f)
        self._d_kd, self._vec, self._omega, self._lum = loads[0:4]
        self.srcviews = loads[4]
        self.srcdir = loads[5]
        try:
            self.srcviewidxs = loads[6]
        except IndexError:
            self.srcviewidxs = [-1] * max(len(self.srcviews), 1)
        self.srcn = self.lum.shape[1]
        f.close()

    def dump(self):
        if self.parent is not None:
            try_mkdir(f"{self.scene.outdir}/{self.parent}")
            try_mkdir(f"{self.scene.outdir}/{self.parent}/{self.src}")
        else:
            try_mkdir(f"{self.scene.outdir}/{self.src}")
        f = open(self.file, 'wb')
        pickle.dump((self._d_kd, self._vec, self._omega, self._lum,
                     self.srcviews, self.srcdir, self.srcviewidxs), f,
                    protocol=4)
        f.close()

    @property
    def vec(self):
        """direction vector (N,3)"""
        return self._vec

    @property
    def lum(self):
        """luminance (N,srcn)"""
        return self._lum

    @property
    def d_kd(self):
        """kd tree for spatial query

        :getter: Returns kd tree structure
        :type: scipy.spatial.cKDTree
        """
        return self._d_kd

    @property
    def omega(self):
        """solid angle (N)

        :getter: Returns array of solid angles
        :setter: sets soolid angles with viewmapper
        :type: np.array
        """
        return self._omega

    def set_srcviews(self, srcviews, idxs=None):
        if srcviews is None:
            srcviews = []
        self.srcviews = [i for i in srcviews
                         if issubclass(type(i), SrcViewPoint) or
                            issubclass(type(i), LightPointKD)]
        if idxs is None or len(idxs) == 0:
            self.srcviewidxs = [-1] * max(len(self.srcviews), 1)
        else:
            self.srcviewidxs = idxs

    def calc_omega(self, write=True):
        """calculate solid angle

        Parameters
        ----------
        write: bool, optional
            update/write kdtree data to file
        """
        vm = self.vm
        if self.vec.shape[0] < 100:
            omega = np.full(len(self.vec), 2 * np.pi * vm.aspect/len(self.vec))
        elif vm.aspect == 1:
            # in case of 180 view, cannot use spherical voronoi, instead
            # the method estimates area in square coordinates by intersecting
            # 2D voronoi with border square.
            # so none of our points have infinite edges.
            uv = vm.xyz2uv(self.vec)
            bordered = np.concatenate((uv, np.array([[-10, -10], [-10, 10],
                                                     [10, 10], [10, -10]])))
            vor = Voronoi(bordered)
            # the border of our 180 degree region
            lb = .5 - vm.viewangle / 360
            ub = .5 + vm.viewangle / 360
            mask = Polygon(np.array([[lb, lb], [lb, ub],
                                     [ub, ub], [ub, lb], [lb, lb]]))
            omega = []
            for i in range(len(uv)):
                region = vor.regions[vor.point_region[i]]
                p = Polygon(np.concatenate((vor.vertices[region],
                                            [vor.vertices[region][
                                                 0]]))).intersection(mask)
                # scaled from unit square -> hemisphere
                omega.append(p.area*2*np.pi)
            omega = np.asarray(omega)
        else:
            try:
                omega = SphericalVoronoi(self.vec).calculate_areas()
            except ValueError:
                # spherical voronoi raises a ValueError when points are
                # too close, in this case we cull duplicates before calculating
                # area, leaving the duplicates with omega=0 it would be more
                # efficient downstream to filter these points, but that would
                # require culling the vecs and lum and rebuilding to kdtree
                omega = np.zeros(self.vec.shape[0])
                tol = 2*np.pi/2**10
                flt = translate.cull_vectors(self.d_kd, tol)
                omega[flt] = SphericalVoronoi(self.vec[flt]).calculate_areas()
        self._omega = omega
        if write:
            self.dump()

    def apply_coef(self, coefs):
        """apply coefficient vector to self.lum

        Parameters
        ----------
        coefs: np.array int float list
            shape (N, self.srcn) or broadcastable

        Returns
        -------
        alum: np.array
            shape (N, self.vec.shape[0])
        """
        try:
            c = np.asarray(coefs).reshape(-1, self.srcn)
        except ValueError:
            c = np.broadcast_to(coefs, (1, self.srcn))
        return np.einsum('ij,kj->ik', c, self.lum)

    def add_to_img(self, img, vecs, mask=None, skyvec=1, interp=False,
                   idx=None, interpweights=None, omega=False, vm=None,
                   rnd=False, engine=None, **kwargs):
        """add luminance contributions to image array (updates in place)

        Parameters
        ----------
        img: np.array
            2D image array to add to (either zeros or with other source)
        vecs: np.array
            vectors corresponding to img pixels shape (N, 3)
        mask: np.array, optional
            indices to img that correspond to vec (in case where whole image
            is not being updated, such as corners of fisheye)
        skyvec: int float np.array, optional
            source coefficients, shape is (1,) or (srcn,)
        interp: Union[bool, str], optional
            if "precomp", use index and interpweights
            if True and engine is None, linearinterpolation
            if "fast" and engine: uses content_interp
            if "high" and engine: uses content_interp_wedge
        idx: np.array, optional
            precomputed query/interpolation result
        interpweights: np.array, optional
            precomputted interpolation weights
        omega: bool
            if true, add value of ray solid angle instead of luminance
        vm: raytraverse.mapper.ViewMapper, optional
        rnd: bool, optional
            use random values as contribution (for visualizing data shape)
        engine: raytraverse.renderer.Rtrace, optional
            engine for content aware interpolation
        kwargs: dict, optional
            passed to interpolationn functions
        """
        if rnd:
            val = np.random.rand(1, self.omega.size)
        elif omega:
            val = self.omega.reshape(1, -1)
        else:
            val = self.apply_coef(skyvec)
        if vm is None:
            vm = self.vm
        if interp == "precomp":
            lum = self.apply_interp(idx, val[0], interpweights)
        elif interp == "fast" and engine is not None:
            i, weights = self.content_interp(engine, vecs, **kwargs)
            lum = self.apply_interp(i, val[0], weights)
        elif interp == "high" and engine is not None:
            i, weights = self.content_interp_wedge(engine, vecs, bandwidth=100,
                                                   **kwargs)
            lum = self.apply_interp(i, val[0], weights)
        elif interp is True:
            lum = self.linear_interp(vm, val[0], vecs)
        elif interp:
            raise ValueError(f"Bad value for interp={interp}, should be boolean"
                             f", 'precomp', 'fast', or 'high'")
        else:
            if idx is not None:
                i = idx
            else:
                i, d = self.query_ray(vecs)
            lum = val[:, i]
        img[mask] += np.squeeze(lum)
        for srcview, srcidx in zip(self.srcviews, self.srcviewidxs):
            srcview.add_to_img(img, vecs, mask, skyvec[srcidx], vm)

    def evaluate(self, skyvec, vm=None, idx=None, srconly=False,
                 blursun=False, includeviews=True):
        """return rays within view with skyvec applied. this is the
        analog to add_to_img for metric calculations

        Parameters
        ----------
        skyvec: int float np.array, optional
            source coefficients, shape is (1,) or (srcn,)
        vm: raytraverse.mapper.ViewMapper, optional
        idx: np.array, optional
            precomputed query_ball result
        srconly: bool, optional
            only evaluate direct sources (stored in self.srcviews)
        includeviews: bool, optional
            include src views in returned results

        Returns
        -------
        rays: np.array
            shape (N, 3) rays falling within view
        omega: np.array
            shape (N,) associated solid angles
        lum: np.array
            shape (N,) associated luminances
        """
        if srconly:
            rays = np.array([[0, 0, 1]])
            omega = np.atleast_1d(np.pi*2)
            lum = np.zeros((1, 1))
        elif vm is None:
            rays = self.vec
            omega = np.atleast_1d(np.squeeze(self.omega))
            lum = self.apply_coef(skyvec).T
            vm = self.vm
        else:
            if idx is None:
                idx = self.query_ball(vm.dxyz, vm.viewangle * vm.aspect)[0]
            omega = np.atleast_1d(np.squeeze(self.omega[idx]))
            rays = self.vec[idx]
            lum = self.apply_coef(skyvec)[:, idx].T
        lum = np.atleast_1d(np.squeeze(lum))
        if len(self.srcviews) > 0 and includeviews:
            vrs = []
            vos = []
            vls = []
            for srcview, srcidx in zip(self.srcviews, self.srcviewidxs):
                srcvec = np.atleast_2d(skyvec)[:, srcidx]
                vr, vo, vl = srcview.evaluate(srcvec, vm, blursun=blursun)
                # 4 times sun
                if vo[0] > 0.00027:
                    cost = np.argsort(-translate.ctheta(vr[0], rays))
                    comega = np.cumsum(omega[cost])
                    cutoff = np.searchsorted(comega, vo[0], 'right')
                    cutoff = max(cutoff-1, 0)
                    cost = cost[cutoff:]
                    rays = rays[cost]
                    lum = lum[cost]
                    omega = omega[cost]
                vrs.append(vr)
                vos.append(vo)
                vls.append(vl)
            rays = np.concatenate([rays] + vrs, axis=0)
            try:
                omega = np.concatenate([omega] + vos, 0)
            except ValueError:
                omega = np.concatenate(vos, axis=0)
            lum = np.concatenate([lum] + vls, axis=0)
        return rays, omega, lum

    def query_ray(self, vecs):
        """return the index and distance of the nearest ray to each of vecs

        Parameters
        ----------
        vecs: np.array
            shape (N, 3) normalized vectors to query, could represent image
            pixels for example.

        Returns
        -------
        i: np.array
            integer indices of closest ray to each query
        d: np.array
            distance (corresponds to chord length on unit sphere) from query to
            ray in lightpoint. use translate.chord2theta to convert to angle.
        """
        d, i = self.d_kd.query(vecs)
        return i, d

    def query_ball(self, vecs, viewangle=180):
        """return set of rays within a view cone

        Parameters
        ----------
        vecs: np.array
            shape (N, 3) vectors to query.
        viewangle: int float
            opening angle of view cone

        Returns
        -------
        i: list np.array
            if vecs is a single point, a list of vector indices of rays
            within view cone. if vecs is a set of point an array of lists, one
            for each vec is returned.
        """
        vs = translate.theta2chord(viewangle/360*np.pi)
        return self.d_kd.query_ball_point(translate.norm(vecs), vs)

    def make_image(self, outf, skyvec, vm=None, res=1024, interp=False,
                   showsample=False):
        if vm is None:
            vm = self.vm
        img, pdirs, mask, mask2, header = vm.init_img(res, self.pt)
        header = [header]
        self.add_to_img(img, pdirs[mask], mask, vm=vm,
                        interp=interp, skyvec=skyvec)
        if showsample:
            img = np.repeat(img[None, ...], 3, 0)
            vi = self.query_ball(vm.dxyz, vm.viewangle * vm.aspect)
            v = self.vec[vi[0]]
            img = vm.add_vecs_to_img(img, v)
            io.carray2hdr(img, outf, header)
        else:
            io.array2hdr(img, outf, header)
        return outf

    def direct_view(self, res=512, showsample=False, showweight=True, rnd=False,
                    srcidx=None, interp=False, omega=False, scalefactor=1,
                    vm=None, fisheye=True):
        """create an unweighted summary image of lightpoint"""
        if vm is None:
            vm = self.vm
        if fisheye:
            img, pdirs, mask, mask2, header = vm.init_img(res, self.pt)
            header = [header]
        else:
            outshape = (res*vm.aspect, res)
            img = np.zeros(outshape)
            uv = translate.bin2uv(np.arange(res*res*vm.aspect), res)
            pdirs = vm.uv2xyz(uv).reshape(res*vm.aspect, res, 3)
            mask = None
            mask2 = None
            header = None
        if showweight:
            if srcidx is not None:
                skyvec = np.zeros(self.srcn)
                skyvec[srcidx] = scalefactor
            else:
                skyvec = np.full(self.srcn, scalefactor)
            self.add_to_img(img, pdirs[mask], mask, vm=vm,
                            interp=interp, skyvec=skyvec, omega=omega, rnd=rnd)
            channels = (1, 0, 0)
        else:
            channels = (1, 1, 1)
        if omega:
            outf = self.file.replace("/", "_").replace(".rytpt", "_omega.hdr")
        else:
            outf = self.file.replace("/", "_").replace(".rytpt", ".hdr")
        if srcidx is not None:
            try:
                outf = outf.replace(f"_{self.src}_", f"_{self.src}_{srcidx:04d}_")
            except TypeError:
                outf = outf.replace(f"_{self.src}_",
                                    f"_{self.src}_filtered_")
        if showsample:
            img = np.repeat(img[None, ...], 3, 0)
            vi = self.query_ball(vm.dxyz, vm.viewangle * vm.aspect)
            v = self.vec[vi[0]]
            img = vm.add_vecs_to_img(img, v, channels=channels,
                                     fisheye=fisheye)
            io.carray2hdr(img, outf, header)
        else:
            io.array2hdr(img, outf, header)
        return outf

    def add(self, lf2, src=None, calcomega=True, write=False, sumsrc=False):
        """add light points of distinct sources together
        results in a new lightpoint with srcn=self.srcn+srcn2 and
        vector size=self.vecsize+vecsize2

        Parameters
        ----------
        lf2: raytraverse.lightpoint.LightPointKD
        src: str, optional
            if None (default), src is "{lf1.src}_{lf2.src}"
        calcomega: bool, optional
            passed to LightPointKD constructor
        write: bool, optional
            passed to LightPointKD constructor
        sumsrc: bool, optional
            if True adds matching source indices together (must be same shape)
            this assumes that the two lightpoints represent the same source
            but different components (such as direct/indirect)
        Returns
        -------
        raytraverse.lightpoint.LightPointKD
            will be subtyped according to self, unless lf2 is needed to preserve
            data
        """
        i, d = self.query_ray(lf2.vec)
        notmatch = d > 1e-6
        i = i[notmatch]
        vecs = np.concatenate((self.vec, lf2.vec[notmatch]), axis=0)

        lum1 = np.concatenate((self.lum, self.lum[i]), axis=0)
        i, d = lf2.query_ray(self.vec)
        lum2 = np.concatenate((lf2.lum[i], lf2.lum[notmatch]), axis=0)
        if sumsrc:
            lums = lum1 + lum2
            srcn = self.srcn
            srcdir = self.srcdir
        else:
            lums = np.concatenate((lum1, lum2), axis=1)
            srcdir = np.concatenate((self.srcdir, lf2.srcdir), axis=0)
            srcn = self.srcn + lf2.srcn
        if src is None:
            src = f"{self.src}_{lf2.src}"
        svi = [i if i >= 0 else self.srcn + i
               for i, sv in zip(self.srcviewidxs, self.srcviews)]
        svi2 = [i + self.srcn if i >= 0 else i
                for i, sv in zip(self.srcviewidxs, self.srcviews)]
        return type(self)(self.scene, vecs, lums, vm=self.vm, pt=self.pt,
                          posidx=self.posidx, src=src, calcomega=calcomega,
                          srcn=srcn, write=write, srcdir=srcdir,
                          srcviews=self.srcviews + lf2.srcviews,
                          srcviewidxs=svi + svi2, filterviews=False,
                          parent=self.parent)

    def update(self, vec, lum, omega=None, calcomega=True, write=True,
               filterviews=False):
        """add additional rays to lightpoint in place

        Parameters
        ----------
        vec: np.array, optional
            shape (N, >=3) where last three columns are normalized direction
            vectors of samples.
        lum: np.array, optional
            reshapeable to (N, srcn). sample values for each source
            corresponding to vec.
        omega: np.array, optional
            provide precomputed omega values, if given, overrides calcomega
        calcomega: bool, optional
            if True (default) calculate solid angle of rays. This  is
            unnecessary if point will be combined before calculating any
            metrics. setting to False will save some computation time. If False,
            resets omega to None!
        write: bool, optional
            whether to save updated ray data to disk.
        filterviews: bool, optional
            delete rays near sourceviews
        """
        self._vec = np.vstack((self.vec, vec[:, -3:]))
        self._lum = np.vstack((self.lum, lum.reshape(-1, self.srcn)))
        self._d_kd = cKDTree(self.vec)
        if filterviews and len(self.srcviews) > 0:
            for sv in self.srcviews:
                lucky_squirel = self.d_kd.query_ball_point(sv.vec, sv.radius)
                if len(lucky_squirel) > 0:
                    self._vec = np.delete(self.vec, lucky_squirel, 0)
                    self._lum = np.delete(self.lum, lucky_squirel, 0)
            self._d_kd = cKDTree(self.vec)
        self._omega = omega
        if calcomega and self._omega is None:
            self.calc_omega(False)
        if write:
            self.dump()

    def linear_interp(self, vm, srcvals, destvecs):
        xyp = vm.xyz2vxy(destvecs)
        xys = vm.xyz2vxy(self.vec)
        interp = LinearNDInterpolator(xys, srcvals, fill_value=-1)
        lum = interp(xyp[:, 0], xyp[:, 1])
        neg = lum < 0
        i, d = self.query_ray(destvecs[neg])
        lum[neg] = srcvals[i]
        return lum

    @staticmethod
    def apply_interp(i, srcvals, weights=None):
        if weights is None:
            return srcvals[i]
        else:
            return np.average(srcvals[i], -1, weights)

    def _c_interp(self, rt, destvecs, bandwidth=10, srfnormtol=0.2,
                  disttol=0.8):

        pts = np.hstack(np.broadcast_arrays(self.pt[None], self.vec))
        pts_o = np.hstack(np.broadcast_arrays(self.pt[None], destvecs))
        d, i = self.d_kd.query(destvecs, bandwidth)
        # store engine state
        targs = rt.args
        tospec = rt.ospec

        rt.set_args(rt.directargs)
        rt.update_ospec("LNM")
        sgeo = rt(pts)
        sgeo_o = rt(pts_o)

        # restore engine state
        rt.set_args(targs)
        rt.update_ospec(tospec)

        mods = sgeo[:, 4].astype(int)
        norms = sgeo[:, 1:4]
        dist = sgeo[:, 0]
        mods_o = sgeo_o[:, 4].astype(int)
        norms_o = sgeo_o[:, 1:4]
        dist_o = sgeo_o[:, 0]

        mod_match = np.equal(mods[i], mods_o[:, None])
        norm_match = np.all(
            np.isclose(norms[i], norms_o[:, None], atol=srfnormtol),
            axis=2)
        dist_match = np.abs(dist[i] - dist_o[:, None]) < disttol
        all_match = np.all((mod_match, norm_match, dist_match), axis=0)
        # make sure atleast the closest match is flagged true
        all_match[np.sum(all_match, 1) == 0, 0] = True
        return all_match, d, i

    def content_interp_wedge(self, rt, destvecs, bandwidth=10, srfnormtol=0.2,
                             disttol=0.8, dp=2, oversample=2, **kwargs):
        all_match, d, i = self._c_interp(rt, destvecs,
                                         bandwidth=bandwidth * oversample,
                                         srfnormtol=srfnormtol, disttol=disttol)
        # reduce to bandwidth prioritizing distance and matching
        no_match = np.logical_not(all_match)
        ds = np.argsort(d + no_match*3)[:, 0:bandwidth]
        no_match = np.take_along_axis(no_match, ds, axis=1)
        d = np.take_along_axis(d, ds, axis=1)
        i = np.take_along_axis(i, ds, axis=1)
        # scale distance as weight
        d = np.power(d, dp)
        d = d[:, 0:1]/d

        dv = destvecs
        sv = self.vec
        # project to interpolation point and calculate theta
        ymtx, pmtx = translate.rmtx_yp(dv)
        nv = np.einsum("vij,vkj,vli->vkl", ymtx, sv[i], pmtx)
        nv[:, 2] = 0
        ang = np.arctan2(nv[..., 1], nv[..., 0]) + np.pi

        # resort by theta
        asr = np.argsort(ang)
        ang = np.take_along_axis(ang, asr, axis=1)
        d = np.take_along_axis(d, asr, axis=1)
        i = np.take_along_axis(i, asr, axis=1)
        no_match = np.take_along_axis(no_match, asr, axis=1)

        # calculate distance matrix between interpolants
        admx = np.abs(ang[:, None] - ang[:, :, None])
        # handle cycle
        admx[admx > np.pi] = 2*np.pi - admx[admx > np.pi]
        # mask out bad candidates
        admx[np.broadcast_to(no_match[:, None], admx.shape)] = 6
        # find the closest alternative as replacement
        replace = np.argmin(admx, 1)
        ir = np.take_along_axis(i, replace, axis=1)
        dr = np.take_along_axis(d, replace, axis=1)
        i[no_match] = ir[no_match]
        d[no_match] = dr[no_match]

        # pad array for circular difference calc
        ang2 = np.hstack((ang[:, -1:] - 2*np.pi, ang, 2*np.pi + ang[:, 0:1]))
        dw = np.hstack((d[:, -1:], d, d[:, 0:1]))

        # apportion angle based on distance
        w1 = (ang - ang2[:, :-2])*d/(d + dw[:, :-2])
        w2 = (ang2[:, 2:] - ang)*d/(d + dw[:, 2:])
        wedge = (w1 + w2)/(2*np.pi)
        return i, wedge

    def content_interp(self, rt, destvecs, bandwidth=10, srfnormtol=0.2,
                       disttol=0.8, **kwargs):
        all_match, d, i = self._c_interp(rt, destvecs, bandwidth=bandwidth,
                                         srfnormtol=srfnormtol, disttol=disttol)
        # weight by the inv. square of distance
        d = 1/np.square(d)
        d[np.logical_not(all_match)] = 0
        # normalize and make cdf based on d
        d = d/np.sum(d, 1)[:, None]
        dc = np.cumsum(d, 1)
        # draw on cdf
        c = np.random.default_rng().random(len(destvecs))
        ic = np.array(list(map(np.searchsorted, dc, c)))[:, None]
        i = np.take_along_axis(i, ic, 1).ravel()
        return i, None
